#   Maybe set up private virtual environment

set( 
    WALBERLA_CODEGEN_PRIVATE_VENV ON
    CACHE BOOL 
    "Create a private virtual Python environment inside the build tree for code generation" 
)

if( WALBERLA_CODEGEN_PRIVATE_VENV )
    set(WALBERLA_CODEGEN_VENV_PATH ${CMAKE_CURRENT_BINARY_DIR}/codegen-venv CACHE PATH "Location of the virtual environment used for code generation")
    set(_venv_python_exe ${WALBERLA_CODEGEN_VENV_PATH}/bin/python)

    find_package( Python COMPONENTS Interpreter REQUIRED )

    if(NOT _sfg_private_venv_done)
        message( STATUS "Setting up Python virtual environment at ${WALBERLA_CODEGEN_VENV_PATH}" )

        #   Create the venv and register its interpreter with pystencils-sfg
        if(NOT EXISTS ${WALBERLA_CODEGEN_VENV_PATH})
            execute_process(
                COMMAND ${Python_EXECUTABLE} -m venv ${WALBERLA_CODEGEN_VENV_PATH}
            )
        endif()

        message( STATUS "Installing required Python packages..." )

        execute_process(
            COMMAND ${_venv_python_exe} -m pip install -r ${sfg_walberla_SOURCE_DIR}/cmake/codegen-requirements.txt
            OUTPUT_QUIET
        )

        execute_process(
            COMMAND ${_venv_python_exe} -m pip install -e ${sfg_walberla_SOURCE_DIR}
            OUTPUT_QUIET
        )

        set( _sfg_private_venv_done TRUE CACHE BOOL "" )
        set( _wlb_codegen_python_init ${_venv_python_exe} )
        mark_as_advanced(_sfg_private_venv_done)
    endif()
else()
    #   Use the external Python environment, but check if all packages are installed
    find_package( Python COMPONENTS Interpreter REQUIRED )

    execute_process(
        COMMAND ${Python_EXECUTABLE} ${CMAKE_CURRENT_SOURCE_DIR}/cmake/check_python_env.py
        RESULT_VARIABLE _package_check_status
        OUTPUT_VARIABLE _package_check_output
        ERROR_VARIABLE _package_check_error
    )

    message( STATUS ${_package_check_output} )

    if(NOT (${_package_check_status} EQUAL 0) )
        message( FATAL_ERROR ${_package_check_error} )
    endif()

    set( _wlb_codegen_python_init ${Python_EXECUTABLE} )
endif()

set(PystencilsSfg_PYTHON_PATH ${_wlb_codegen_python_init})
set(
    WALBERLA_CODEGEN_PYTHON
    ${_wlb_codegen_python_init}
    CACHE PATH
    "Path to the Python interpreter used for code generation"
)
mark_as_advanced(WALBERLA_CODEGEN_PYTHON)

#   Find pystencils-sfg

find_package( PystencilsSfg REQUIRED )

#   Project Configuration Module

set( 
    WALBERLA_CODEGEN_CONFIG_MODULE
    ${CMAKE_BINARY_DIR}/CodegenConfig.py
    CACHE
    FILEPATH
    "Path to waLBerla-wide codegen config module" 
)
mark_as_advanced( WALBERLA_CODEGEN_CONFIG_MODULE )

configure_file(
    ${CMAKE_CURRENT_SOURCE_DIR}/cmake/CodegenConfig.template.py
    ${WALBERLA_CODEGEN_CONFIG_MODULE}
)

message( STATUS "Wrote project-wide code generator configuration to ${WALBERLA_CODEGEN_CONFIG_MODULE}" )

#[[
Run `pip install` in the code generation environment with the given arguments.
#]]
function(walberla_codegen_venv_install)
    if(NOT WALBERLA_CODEGEN_PRIVATE_VENV)
        return()
    endif()

    if(NOT _sfg_private_venv_done)
        message( FATAL_ERROR "The private virtual environment for code generation was not initialized yet" )
    endif()

    execute_process(
        COMMAND ${WALBERLA_CODEGEN_PYTHON} -m pip install ${ARGV}
    )
endfunction()

#   Code Generation Functions

#[[
Register code generation scripts for a CMake target.

Signature:

```
walberla_generate_sources( <target> 
    SCRIPTS script1.py [script2.py ...]
    [DEPENDS dependency1.py [dependency2.py...] ]
    [FILE_EXTENSIONS <header-extension> <impl-extension>]
    [OUTPUT_MODE <standalone|inline|header-only>]
)
```

This is a wrapper around `pystencilssfg_generate_target_sources`
without the `CONFIG_MODULE` parameter.
See also https://pycodegen.pages.i10git.cs.fau.de/pystencils-sfg/usage/project_integration.html#add-generator-scripts
#]]
function(walberla_generate_sources TARGET)
    pystencilssfg_generate_target_sources(${ARGV} CONFIG_MODULE $CACHE{WALBERLA_CODEGEN_CONFIG_MODULE})
endfunction()
