#pragma once

#include "blockforest/StructuredBlockForest.h"

#include "core/DataTypes.h"
#include "core/cell/Cell.h"

#include "domain_decomposition/BlockDataID.h"

#include "field/FlagField.h"

#include "stencil/Directions.h"

#include <memory>
#include <tuple>

#include "SparseIteration.hpp"

namespace walberla
{
  namespace sfg
  {

    struct HbbLink
    {
      int32_t x;
      int32_t y;
      int32_t z;
      int32_t dir;

      /**
       * Create a halfway bounce-back boundary link from a fluid cell `c` to a boundary cell in direction `d`.
       */
      HbbLink(const Cell &c, const stencil::Direction d)
          : x{int32_c(c.x())}, y{int32_c(c.y())}, z{int32_c(c.z())}, dir{int32_t(d)}
      {
      }

      bool operator==(const HbbLink &other) const
      {
        return std::tie(x, y, z, dir) == std::tie(other.x, other.y, other.z, other.dir);
      }
    };

    /**
     * CRTP base class for factories of generated half-way bounce back boundary conditions.
     */
    template <typename Impl>
    class GenericHbbFactory
    {
    public:
      GenericHbbFactory(const shared_ptr<StructuredBlockForest> blocks) : blocks_{blocks} {}

      template <typename FlagField_T>
      auto fromFlagField(BlockDataID flagFieldID, field::FlagUID boundaryFlagUID, field::FlagUID fluidFlagUID)
      {
        using Stencil = typename Impl::Stencil;
        using IndexList = SparseIndexList<HbbLink>;

        IndexList indexList{blocks_};

        for (auto &block : blocks_)
        {
          FlagField_T *flagField = block.getData<FlagField_T>(flagFieldID_);
          flag_t boundaryFlag{flagField->getFlag(boundaryFlagUID)};
          flag_t fluidFlag{flagField->getFlag(fluidFlagUID)};

          auto &idxVector = indexList.get(block);

          for (auto it = flagField->beginXYZ(); it != flagField->end(); ++it)
          {
            Cell c{it.cell()};
            if (!isFlagSet(it, fluidFlag))
            {
              continue;
            }

            for (auto dIt = Stencil::beginNoCenter(); dIt != Stencil::end(); ++dIt)
            {
              stencil::Direction dir{*dIt};
              Cell neighbor{c + dir};
              if (flagField->isFlagSet(neighbor, boundaryFlag))
              {
                indexVectorAll.emplace_back(fluidCell, dir);
              }
            }
          }
        }

        return indexList;

        return impl().irregularFromIndexVector(indexVector);
      }

    private:
      shared_ptr<StructuredBlockForest> blocks_;
    };

  } // namespace sfg
} // namespace walberla
