#pragma once

#include <vector>
#include "core/DataTypes.h"
#include "core/cell/Cell.h"

#include "domain_decomposition/BlockDataID.h"
#include "blockforest/StructuredBlockForest.h"

namespace walberla::sfg
{
  namespace internal
  {
    /**
     * @brief Index vectors for sparse sweeps
     */
    template <typename IndexStruct>
    class IndexListBuffer
    {
    public:
      using CpuVector = std::vector<IndexStruct>;

      IndexListBuffer() = default;
      bool operator==(const IndexListBuffer &other) const { return other.vec_ == vec_; }

      CpuVector &vector() { return vec_; }

      IndexStruct *pointerCpu() { return vec_.data(); }

    private:
      CpuVector vec_;
    };
  }

  struct CellIdx
  {
    int64_t x;
    int64_t y;
    int64_t z;

    CellIdx(const Cell &c)
        : x{int64_c(c.x())}, y{int64_c(c.y())}, z{int64_c(c.z())}
    {
    }

    bool operator==(const CellIdx &other) const
    {
      return std::tie(x, y, z) == std::tie(other.x, other.y, other.z);
    }
  };

  template <typename IndexStruct = CellIdx>
  class SparseIndexList
  {
  public:
    using Buffer_T = internal::IndexListBuffer<IndexStruct>;

    explicit SparseIndexList(StructuredBlockForest &sbfs) : bufferId_{createBuffers(sbfs)} {}

    std::vector<IndexStruct> &get(IBlock &block)
    {
      Buffer_T *buf = block.template getData<Buffer_T>(bufferId_);
      return buf->vector();
    }

    const BlockDataID & bufferId() const {
      return bufferId_;
    }

  private:
    static BlockDataID createBuffers(StructuredBlockForest &sbfs)
    {
      auto createIdxVector = [](IBlock *const, StructuredBlockStorage *const)
      { return new Buffer_T(); };
      return sbfs.addStructuredBlockData<Buffer_T>(createIdxVector);
    }

    BlockDataID bufferId_;
  };
}
