from __future__ import annotations

from typing import cast

from pystencils import Field
from pystencils.types import (
    UserTypeSpec,
    create_type,
    PsCustomType,
    PsPointerType,
    PsType,
    PsStructType,
)
from pystencilssfg.lang import (
    IFieldExtraction,
    AugExpr,
    SrcField,
    SrcVector,
    Ref,
    ExprLike,
    cpptype,
    CppClass,
)
from pystencilssfg.lang.types import CppTypeFactory, CppType
from pystencilssfg.lang.cpp import std


real_t = PsCustomType("walberla::real_t")
cell_idx_t = PsCustomType("walberla::cell_idx_t")
uint_t = PsCustomType("walberla::uint_t")


class _PlainCppClass(AugExpr):
    _type: CppTypeFactory

    def __init__(self, const: bool = False, ref: bool = False):
        dtype = self._type(const=const, ref=ref)
        super().__init__(dtype)


class Vector2(SrcVector):
    _template = cpptype("walberla::Vector2< {element_type} >", "core/math/Vector2.h")

    def __init__(
        self, element_type: UserTypeSpec, const: bool = False, ref: bool = False
    ):
        self._element_type = create_type(element_type)
        dtype = self._template(element_type=element_type, const=const, ref=ref)
        super().__init__(dtype)

    def extract_component(self, coordinate: int) -> AugExpr:
        if coordinate > 1:
            raise ValueError(f"Cannot extract component {coordinate} from Vector2")

        return AugExpr(self._element_type).bind("{}[{}]", self, coordinate)


class Vector3(SrcVector):
    _template = cpptype("walberla::Vector3< {element_type} >", "core/math/Vector3.h")

    def __init__(
        self, element_type: UserTypeSpec, const: bool = False, ref: bool = False
    ):
        self._element_type = create_type(element_type)
        dtype = self._template(element_type=element_type, const=const, ref=ref)
        super().__init__(dtype)

    def extract_component(self, coordinate: int) -> AugExpr:
        if coordinate > 2:
            raise ValueError(f"Cannot extract component {coordinate} from Vector3")

        return AugExpr(self._element_type).bind("{}[{}]", self, coordinate)

    def __getitem__(self, idx: int | ExprLike):
        return AugExpr(self._element_type).bind("{}[{}]", self, idx)


class AABB(_PlainCppClass):
    _type = cpptype("walberla::AABB", "core/math/AABB.h")

    def min(self) -> Vector3:
        return Vector3(real_t).bind("{}.min()", self)

    def max(self) -> Vector3:
        return Vector3(real_t).bind("{}.max()", self)


class Cell(_PlainCppClass):
    _type = cpptype("walberla::Cell", "core/cell/Cell.h")

    def x(self) -> AugExpr:
        return AugExpr.format("{}.x()", self)

    def y(self) -> AugExpr:
        return AugExpr.format("{}.y()", self)

    def z(self) -> AugExpr:
        return AugExpr.format("{}.z()", self)


class CellInterval(_PlainCppClass):
    _type = cpptype("walberla::CellInterval", "core/cell/CellInterval.h")


class Direction(_PlainCppClass):
    _type = cpptype("walberla::stencil::Direction", "stencil/Directions.h")


class BlockDataID(_PlainCppClass):
    _type = cpptype("walberla::BlockDataID", "domain_decomposition/BlockDataID.h")


class IBlock(_PlainCppClass):
    _type = cpptype("walberla::IBlock", "domain_decomposition/IBlock.h")

    def getData(self, dtype: str | PsType, id: BlockDataID) -> AugExpr:
        return AugExpr.format("{}.template getData< {} >({})", self, dtype, id)

    def getAABB(self) -> AABB:
        return AABB().bind("{}.getAABB()", self)


class IBlockPtr(_PlainCppClass):
    _type = cpptype("walberla::IBlock *", "domain_decomposition/IBlock.h")

    def getData(self, dtype: str | PsType, id: BlockDataID) -> AugExpr:
        return AugExpr.format("{}->template getData< {} >({})", self, dtype, id)

    def getAABB(self) -> AABB:
        return AABB().bind("{}->getAABB()", self)

    def deref(self) -> AugExpr:
        return AugExpr.format("*{}", self)


class SharedPtr(CppClass):
    template = cpptype("std::shared_ptr< {} >", "<memory>")


class StructuredBlockForest(AugExpr):
    typename = cpptype(
        "walberla::StructuredBlockForest", "blockforest/StructuredBlockForest.h"
    )

    def __init__(self, ref: bool = True, const: bool = False):
        dtype = self.typename(const=const, ref=ref)
        super().__init__(dtype)

    @staticmethod
    def shared_ptr(name: str = "blocks", const: bool = False):
        dtype = StructuredBlockForest.typename(const=const)
        return SharedPtr(dtype).var(name)

    @staticmethod
    def shared_ptr_ref(name: str = "blocks", const: bool = False):
        dtype = StructuredBlockForest.typename(const=const)
        return SharedPtr(dtype, const=True, ref=True).var(name)

    @staticmethod
    def ref(name: str = "blocks", const: bool = False):
        return StructuredBlockForest(ref=True).var(name)

    def getDomain(self) -> AABB:
        return AABB().bind("{}.getDomain()", self)

    def dx(self, level: int | AugExpr = 0) -> AugExpr:
        return AugExpr(real_t).bind("{}.dx({})", self, level)

    def dy(self, level: int | AugExpr = 0) -> AugExpr:
        return AugExpr(real_t).bind("{}.dy({})", self, level)

    def dz(self, level: int | AugExpr = 0) -> AugExpr:
        return AugExpr(real_t).bind("{}.dz({})", self, level)

    def cell_extents(self, coord: int, level: int | AugExpr = 0) -> AugExpr:
        dx_getter = [self.dx, self.dy, self.dz][coord]
        return dx_getter(level)

    def cellsPerBlock(self, coord: int, block: AugExpr) -> AugExpr:
        x = ["X", "Y", "Z"][coord]
        return AugExpr(uint_t).bind("{}.getNumberOf{}Cells({})", self, x, block)

    def getLevel(self, block: AugExpr) -> AugExpr:
        return AugExpr(uint_t).bind("{}.getLevel({})", self, block)


class GenericWalberlaField(SrcField):
    """Common base class for GhostLayerField and GpuField defining their shared interface."""

    def __init__(
        self,
        element_type: PsType,
        field_type: PsCustomType,
        ptr: bool = False,
        ref: bool = False,
    ):
        if ptr and ref:
            raise ValueError("At most one of `ptr` and `ref` may be true.")

        self._element_type = element_type
        self._field_type = field_type

        if ptr:
            obj_type = PsPointerType(self._field_type)
        elif ref:
            obj_type = Ref(self._field_type)
        else:
            obj_type = self._field_type

        super().__init__(obj_type)

    @property
    def _a(self) -> AugExpr:
        """Member access"""
        if isinstance(self.dtype, PsPointerType):
            return AugExpr.format("{}->", self)
        else:
            return AugExpr.format("{}.", self)

    @property
    def field_type(self) -> PsCustomType:
        return self._field_type

    def get_extraction(self) -> IFieldExtraction:
        return GhostLayerFieldExtraction(self, None)

    def with_cell_interval(self, ci: CellInterval | None):
        return GhostLayerFieldExtraction(self, ci)

    def cloneUninitialized(self) -> AugExpr:
        return AugExpr.format("{}cloneUninitialized()", self._a)

    def swapDataPointers(self, other: AugExpr) -> AugExpr:
        return AugExpr.format("{}swapDataPointers({});", self._a, other)


class GhostLayerFieldPtr(GenericWalberlaField):
    _template = cpptype(
        "walberla::field::GhostLayerField< {element_type}, {fsize} >",
        "field/GhostLayerField.h",
    )

    @staticmethod
    def create(field: Field):
        if field.index_dimensions > 1:
            raise ValueError(
                "Cannot map fields with more than one index dimension to field::GhostLayerField."
            )

        element_type = field.dtype
        fsize = field.index_shape[0] if field.index_shape else 1

        return GhostLayerFieldPtr(element_type, fsize).var(field.name)

    def __init__(
        self,
        element_type: UserTypeSpec,
        fsize: int,
    ):
        element_type = create_type(element_type)
        field_type = self._template(element_type=element_type, fsize=fsize)

        super().__init__(element_type, field_type, ptr=True)


class GpuFieldPtr(GenericWalberlaField):
    _template = cpptype(
        "walberla::gpu::GpuField< {element_type} >",
        "gpu/GpuField.h",
    )

    @staticmethod
    def create(field: Field):
        if field.index_dimensions > 1:
            raise ValueError(
                "Cannot map fields with more than one index dimension to gpu::GpuField."
            )

        element_type = field.dtype
        fsize = field.index_shape[0] if field.index_shape else 1

        return GpuFieldPtr(element_type, fsize).var(field.name)

    def __init__(
        self,
        element_type: UserTypeSpec,
        fsize: int,
    ):
        element_type = create_type(element_type)
        field_type = self._template(element_type=element_type)

        super().__init__(element_type, field_type, ptr=True)


class GhostLayerFieldExtraction(IFieldExtraction):
    def __init__(
        self,
        field_ptr: AugExpr,
        cell_interval: AugExpr | None = None,
    ) -> None:
        self._field_ptr = field_ptr
        self._ci = cell_interval

    def ptr(self) -> AugExpr:
        data_at: AugExpr | str
        if self._ci is not None:
            ci = self._ci
            data_at = AugExpr.format("{ci}.xMin(), {ci}.yMin(), {ci}.zMin(), 0", ci=ci)
        else:
            data_at = "0, 0, 0, 0"

        return AugExpr.format("{}->dataAt({})", self._field_ptr, data_at)

    def size(self, coordinate: int) -> AugExpr | None:
        size_call = ["xSize()", "ySize()", "zSize()", "fSize()"][coordinate]

        if self._ci is not None and coordinate < 3:
            return AugExpr.format("{}.{}", self._ci, size_call)
        else:
            return AugExpr.format("{}->{}", self._field_ptr, size_call)

    def stride(self, coordinate: int) -> AugExpr | None:
        stride_call = ("xStride()", "yStride()", "zStride()", "fStride()")[coordinate]
        return AugExpr.format("{}->{}", self._field_ptr, stride_call)


def glfield(field: Field, ci: str | AugExpr | None = None):
    field_ptr = GhostLayerFieldPtr.create(field)

    if ci is not None and not isinstance(ci, AugExpr):
        ci = AugExpr(PsCustomType("walberla::CellInterval", const=True)).var(ci)

    return GhostLayerFieldExtraction(field_ptr, ci)


class SparseIndexList(AugExpr):
    _template = cpptype(
        "walberla::sfg::SparseIndexList< {IndexStruct} >", "sfg/SparseIteration.hpp"
    )

    def __init__(
        self, idx_struct: PsStructType, const: bool = False, ref: bool = False
    ):
        self._idx_struct = idx_struct
        dtype = self._template(IndexStruct=idx_struct, const=const, ref=ref)
        super().__init__(dtype)

    def get(self, block: IBlock) -> std.vector:
        return std.vector(self._idx_struct, ref=True).bind("{}.get({})", self, block)

    def bufferId(self) -> BlockDataID:
        return BlockDataID().bind("{}.bufferId()", self)


class IndexListBufferPtr(SrcField):
    _template = cpptype(
        "walberla::sfg::internal::IndexListBuffer< {IndexStruct} >",
        "sfg/SparseIteration.hpp",
    )

    def __init__(self, idx_struct: PsStructType):
        self._idx_struct = idx_struct
        dtype = self._template(IndexStruct=idx_struct)
        self._base_type = dtype
        super().__init__(PsPointerType(dtype))

    @property
    def field_type(self) -> CppType:
        return cast(CppType, self._base_type)

    def pointerCpu(self) -> AugExpr:
        return AugExpr().format("{}->pointerCpu()", self)

    def vector(self) -> AugExpr:
        return AugExpr().format("{}->vector()", self)

    def get_extraction(self) -> IFieldExtraction:
        class IdxVectorExtraction(IFieldExtraction):
            def ptr(_) -> AugExpr:
                return self.pointerCpu()

            def size(_, coordinate: int) -> AugExpr | None:
                if coordinate == 0:
                    return AugExpr.format("{}.size()", self.vector())
                elif coordinate == 1:
                    return AugExpr.format("1")
                else:
                    raise ValueError()

            def stride(_, coordinate: int) -> AugExpr | None:
                if coordinate <= 1:
                    return AugExpr.format("1")
                else:
                    raise ValueError()

        return IdxVectorExtraction()


CellIdx = PsStructType(
    (
        ("x", create_type("int64_t")),
        ("y", create_type("int64_t")),
        ("z", create_type("int64_t")),
    ),
    "walberla::sfg::CellIdx",
)
