from functools import cache

from pystencils import Field, FieldType, TypedSymbol, Assignment
from pystencils.types import PsStructType, create_type

from lbmpy import LBStencil
from lbmpy.methods import AbstractLbMethod
from lbmpy.boundaries.boundaryconditions import LbBoundary
from lbmpy.advanced_streaming import Timestep

from pystencilssfg import SfgComposer
from pystencilssfg.composer.custom import CustomGenerator

BoundaryIndexType = create_type("int32")

HbbLinkType = PsStructType(
    [
        ("x", BoundaryIndexType),
        ("y", BoundaryIndexType),
        ("z", BoundaryIndexType),
        ("dir", BoundaryIndexType),
    ],
    "walberla::sfg::HbbLink",
)


class WalberlaLbmBoundary:
    idx_struct_type: PsStructType

    @classmethod
    @cache
    def get_index_field(cls):
        return Field(
            "indexVector",
            FieldType.INDEXED,
            cls.idx_struct_type,
            (0,),
            (TypedSymbol("indexVectorLength", BoundaryIndexType), 1),
            (1, 1),
        )

    def get_assignments(
        self,
        lb_method: AbstractLbMethod,
        pdf_field: Field,
        prev_timestep: Timestep = Timestep.BOTH,
        streaming_pattern="pull",
    ) -> list[Assignment]:
        assert isinstance(self, LbBoundary)

        index_field = self.get_index_field()

        from lbmpy.advanced_streaming.indexing import BetweenTimestepsIndexing

        indexing = BetweenTimestepsIndexing(
            pdf_field,
            lb_method.stencil,
            prev_timestep,
            streaming_pattern,
            BoundaryIndexType,
            BoundaryIndexType,
        )

        f_out, f_in = indexing.proxy_fields
        dir_symbol = indexing.dir_symbol
        inv_dir = indexing.inverse_dir_symbol

        boundary_assignments = self(
            f_out,
            f_in,
            dir_symbol,
            inv_dir,
            lb_method,
            index_field,
            None,  # TODO: Fix force vector
        )
        boundary_assignments = indexing.substitute_proxies(boundary_assignments)

        elements: list[Assignment] = []

        index_arrs_node = indexing.create_code_node()
        elements += index_arrs_node.get_array_declarations()

        for node in self.get_additional_code_nodes(lb_method)[::-1]:
            elements += node.get_array_declarations()

        elements += [Assignment(dir_symbol, index_field[0]("dir"))]
        elements += boundary_assignments.all_assignments

        return elements
