from pystencils import Field, Assignment, CreateKernelConfig
from pystencils.types import PsStructType

from lbmpy.methods import AbstractLbMethod
from lbmpy.boundaries.boundaryconditions import LbBoundary

from pystencilssfg import SfgComposer
from pystencilssfg.composer.custom import CustomGenerator

from .hbb import BoundaryIndexType
from .boundary_utils import WalberlaLbmBoundary
from ..sweep import Sweep
from ..api import SparseIndexList

__all__ = ["FreeSlip"]


class _IrregularSentinel:
    def __repr__(self) -> str:
        return "IRREGULAR"


class FreeSlip(CustomGenerator):
    """Free-Slip boundary condition"""

    IRREGULAR = _IrregularSentinel()

    def __init__(
        self,
        name: str,
        lb_method: AbstractLbMethod,
        pdf_field: Field,
        wall_normal: tuple[int, int, int] | _IrregularSentinel,
    ):
        self._name = name
        self._method = lb_method
        self._pdf_field = pdf_field
        self._wall_normal = wall_normal

    def generate(self, sfg: SfgComposer) -> None:
        if self._wall_normal == self.IRREGULAR:
            self._generate_irregular(sfg)
        else:
            self._generate_regular(sfg)

    def _generate_irregular(self, sfg: SfgComposer):
        sfg.include("sfg/IrregularFreeSlip.hpp")

        #   Get waLBerla build config
        bc_obj = WalberlaIrregularFreeSlip()

        #   Get assignments for bc
        bc_asm = bc_obj.get_assignments(self._method, self._pdf_field)

        #   Build generator config
        bc_cfg = CreateKernelConfig()
        bc_cfg.index_dtype = BoundaryIndexType
        index_field = bc_obj.get_index_field()
        bc_cfg.index_field = index_field

        #   Prepare sweep
        bc_sweep = Sweep(self._name, bc_asm, bc_cfg)

        #   Emit code
        sfg.generate(bc_sweep)

        #   Build factory
        factory_name = f"{self._name}Factory"
        factory_crtp_base = f"walberla::sfg::IrregularFreeSlipFactory< {factory_name} >"
        index_vector = SparseIndexList(
            WalberlaIrregularFreeSlip.idx_struct_type, ref=True
        ).var("indexVector")

        sweep_type = bc_sweep.generated_class()
        sweep_ctor_args = {
            f"{self._pdf_field.name}Id": "this->pdfFieldID_",
            f"{index_field.name}Id": index_vector.bufferId(),
        }

        stencil_name = self._method.stencil.name
        sfg.include(f"stencil/{stencil_name}.h")

        sfg.klass(factory_name, bases=[f"public {factory_crtp_base}"])(
            sfg.public(
                f"using Base = {factory_crtp_base};",
                f"friend class {factory_crtp_base};",
                f"using Stencil = walberla::stencil::{stencil_name};",
                f"using Sweep = {sweep_type.get_dtype().c_string()};",
                "using Base::Base;",
            ),
            sfg.private(
                sfg.method(
                    "irregularFromIndexVector",
                    returns=sweep_type.get_dtype(),
                    inline=True,
                )(sfg.expr("return {};", sweep_type.ctor(**sweep_ctor_args))),
            ),
        )

    def _generate_regular(self, sfg: SfgComposer):
        raise NotImplementedError("Regular-geometry free slip is not implemented yet")


class WalberlaIrregularFreeSlip(LbBoundary, WalberlaLbmBoundary):
    idx_struct_type = PsStructType(
        (
            ("x", BoundaryIndexType),
            ("y", BoundaryIndexType),
            ("z", BoundaryIndexType),
            ("dir", BoundaryIndexType),
            ("source_offset_x", BoundaryIndexType),
            ("source_offset_y", BoundaryIndexType),
            ("source_offset_z", BoundaryIndexType),
            ("source_dir", BoundaryIndexType),
        ),
        "walberla::sfg::IrregularFreeSlipLinkInfo",
    )

    def __call__(
        self, f_out, f_in, dir_symbol, inv_dir, lb_method, index_field, force_vector
    ):
        source_cell = (
            index_field("source_offset_x"),
            index_field("source_offset_y"),
            index_field("source_offset_z"),
        )
        source_dir = index_field("source_dir")

        return Assignment(f_in(inv_dir[dir_symbol]), f_out[source_cell](source_dir))
