from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Sequence
from functools import cache

from pystencilssfg import SfgComposer
from pystencilssfg.composer.custom import CustomGenerator
from pystencilssfg.lang import HeaderFile
from pystencilssfg.lang.types import CppTypeFactory

from pystencils import Assignment, CreateKernelConfig, Field, FieldType, TypedSymbol
from pystencils.types import PsStructType

from lbmpy import LBStencil

from ..build_config import WalberlaBuildConfig
from .boundary_utils import BoundaryIndexType
from ..sweep import Sweep
from ..api import SparseIndexList, 


class _IrregularSentinel:
    def __repr__(self) -> str:
        return "IRREGULAR"


class GenericLbmBoundary(CustomGenerator, ABC):

    IRREGULAR = _IrregularSentinel()

    #   Subclasses must-defines

    idx_struct_type: PsStructType
    includes: Sequence[HeaderFile]
    factory_base: CppTypeFactory

    @abstractmethod
    def _get_sparse_assignments(self) -> list[Assignment]: ...

    #   Utility

    @classmethod
    @cache
    def get_index_field(cls):
        return Field(
            "indexVector",
            FieldType.INDEXED,
            cls.idx_struct_type,
            (0,),
            (TypedSymbol("indexVectorLength", BoundaryIndexType), 1),
            (1, 1),
        )

    #   Code Generation Protocols

    def __init__(
        self,
        name: str,
        stencil: LBStencil,
        wall_normal: tuple[int, int, int] | _IrregularSentinel,
        gen_config: CreateKernelConfig
    ):
        self._name = name
        self._stencil = stencil
        self._wall_normal = wall_normal
        self._user_gen_config = gen_config

    def generate(self, sfg: SfgComposer) -> None:
        if self._wall_normal == self.IRREGULAR:
            self._generate_irregular(sfg)
        else:
            self._generate_regular(sfg)

    def _generate_irregular(self, sfg: SfgComposer):
        for header in self.includes:
            sfg.include(header)

        #   Get assignments for bc
        bc_asm = self._get_sparse_assignments()

        #   Build generator config
        bc_cfg = WalberlaBuildConfig.from_sfg(sfg).get_pystencils_config()
        bc_cfg.override(self._user_gen_config)
        bc_cfg.index_dtype = BoundaryIndexType
        index_field = self.get_index_field()
        bc_cfg.index_field = index_field

        #   Prepare sweep
        bc_sweep = Sweep(self._name, bc_asm, bc_cfg)

        #   Emit code
        sfg.generate(bc_sweep)

        #   Build factory
        factory_name = f"{self._name}Factory"
        factory_crtp_base = f"walberla::sfg::IrregularFreeSlipFactory< {factory_name} >"
        index_vector = SparseIndexList(
            self.idx_struct_type, ref=True
        ).var("indexVector")

        sweep_type = bc_sweep.generated_class()

        stencil_name = self._stencil.name
        sfg.include(f"stencil/{stencil_name}.h")

        #   TODO: To forward sweep properties through the factory,
        #   we need to encapsulate them in a PropertyCache object!
        #   Otherwise, this will get quite messy.

        sfg.klass(factory_name, bases=[f"public {factory_crtp_base}"])(
            sfg.public(
                f"using Base = {factory_crtp_base};",
                f"friend class {factory_crtp_base};",
                f"using Stencil = walberla::stencil::{stencil_name};",
                f"using Sweep = {sweep_type.get_dtype().c_string()};",
                "using Base::Base;",
            ),
            sfg.private(
                sfg.method(
                    "irregularFromIndexVector",
                    returns=sweep_type.get_dtype(),
                    inline=True,
                )(sfg.expr("return {};", sweep_type.ctor(**sweep_ctor_args))),
            ),
        )

    def _generate_regular(self, sfg: SfgComposer):
        raise NotImplementedError("Regular-geometry free slip is not implemented yet")
