"""Code generation for half-way bounce-back LBM boundary conditions in waLBerla"""

from pystencils import Field, FieldType, TypedSymbol, Target, AssignmentCollection, DynamicType

from pystencils.types import PsStructType, create_type

from pystencilssfg import SfgComposer
from pystencilssfg.composer.class_composer import SfgClassComposer
from pystencilssfg.composer.custom import CustomGenerator
from pystencilssfg.lang import AugExpr

from lbmpy import LBStencil
from lbmpy.methods import AbstractLbMethod
from lbmpy.boundaries.boundaryconditions import LbBoundary
from lbmpy.advanced_streaming.indexing import Timestep

from ..sweep import (
    FieldInfo,
    SweepClassProperties,
    BlockforestParameters,
    combine_vectors,
)
from ..api import GhostLayerFieldPtr, BlockDataID, IBlockPtr, StructuredBlockForest, IndexListBufferPtr
from .boundary_utils import HbbLinkType, BoundaryIndexType


class HbbBoundaryProperties(SweepClassProperties):
    def _constructor(self, sfg: SfgComposer) -> SfgClassComposer.ConstructorBuilder:
        blocks = StructuredBlockForest.shared_ptr_ref()
        return sfg.constructor(blocks).init("GenericHbbBoundary")(
            AugExpr.format("*{}", blocks)
        )


class SimpleHbbBoundary(CustomGenerator):
    def __init__(self, bc: LbBoundary, lb_method: AbstractLbMethod, pdf_field: Field):
        if bc.additional_data:
            raise ValueError(
                "The SimpleHbbBoundary generator only supports boundary conditions without "
                "additional per-link data."
            )

        self._bc = bc
        self._name = bc.name
        self._lb_method = lb_method
        self._pdf_field = pdf_field
        self._index_field = self._get_index_field()
        self._stencil = lb_method.stencil

    def generate(self, sfg: SfgComposer) -> None:
        sfg.include("sfg/GenericHbbBoundary.hpp")
        sfg.include(f"stencil/{self._stencil.name}.h")

        knamespace = sfg.kernel_namespace(f"{self._name}_kernels")
        bc_kernel = _create_lb_boundary_kernel(
            self._pdf_field, self._get_index_field(), self._lb_method, self._bc
        )

        khandle = knamespace.add(bc_kernel, self._name)

        domain_fields: list[FieldInfo] = sorted(
            (
                FieldInfo(
                    f, GhostLayerFieldPtr.create(f), BlockDataID().var(f"{f.name}Id")
                )
                for f in khandle.fields
                if f.field_type == FieldType.GENERIC
            ),
            key=lambda f: f.field.name,
        )

        props = HbbBoundaryProperties()
        block = IBlockPtr().var("block")

        parameters = khandle.scalar_parameters

        blockforest_params = BlockforestParameters(props, block, None)
        parameters = blockforest_params.filter_params(parameters)

        vector_groups = combine_vectors(parameters)

        for fi in domain_fields:
            props.add_property(fi.data_id, setter=False, getter=True)

        for s in sorted(parameters, key=lambda p: p.name):
            props.add_property(s, setter=True, getter=True)

        idx_vector = IndexListBufferPtr(HbbLinkType).var("indexVector")
        idx_vector_id = BlockDataID().bind("this->indexVectorsId_")

        base_class = f"walberla::sfg::GenericHbbBoundary< walberla::stencil::{self._stencil.name} >"

        sfg.klass(self._name, bases=[f"public {base_class}"])(
            *props.render(sfg),
            sfg.public(
                sfg.method("operator()")(
                    #   Get block data IDs for fields
                    *(
                        sfg.init(fi.data_id)(props.get(fi.data_id))
                        for fi in domain_fields
                    ),
                    #   Get fields from block
                    *(
                        sfg.init(fi.wlb_field_ptr)(
                            block.getData(fi.wlb_field_ptr.field_type, fi.data_id)
                        )
                        for fi in domain_fields
                    ),
                    #   Map GhostLayerFields to pystencils fields
                    *(sfg.map_field(fi.field, fi.wlb_field_ptr) for fi in domain_fields),
                    #   Get index vector
                    sfg.init(idx_vector)(
                        block.getData(idx_vector.field_type, idx_vector_id)
                    ),
                    #   Map index vector to pystencils field
                    sfg.map_field(self._index_field, idx_vector),
                    #   Extract parameters
                    *(sfg.init(param)(props.get(param)) for param in parameters),
                    #   Extract vector components
                    *(
                        sfg.map_vector(comps, vec)
                        for vec, comps in vector_groups.items()
                    ),
                    #   Extract geometry information
                    *(blockforest_params.render_extractions(sfg)),
                    #   Call kernel
                    sfg.call(khandle),
                )
            ),
        )

    def _get_index_field(self) -> Field:
        return Field(
            "indexField",
            FieldType.INDEXED,
            HbbLinkType,
            (0,),
            (TypedSymbol("index_vector_length", BoundaryIndexType), 1),
            (1, 1),
        )


def _create_lb_boundary_kernel(
    pdf_field,
    index_field,
    lb_method,
    boundary_functor,
    prev_timestep=Timestep.BOTH,
    streaming_pattern="pull",
    target=Target.CPU,
    **kernel_creation_args,
):
    """Almost-copy of `lbmpy.boundaryhandling.create_lattice_boltzmann_boundary_kernel`,
    extended with waLBerla-specific function expansion."""

    from lbmpy.advanced_streaming.indexing import BetweenTimestepsIndexing

    indexing = BetweenTimestepsIndexing(
        pdf_field, lb_method.stencil, prev_timestep, streaming_pattern, "int32", "int32"
    )

    f_out, f_in = indexing.proxy_fields
    dir_symbol = indexing.dir_symbol
    inv_dir = indexing.inverse_dir_symbol

    boundary_assignments = boundary_functor(
        f_out, f_in, dir_symbol, inv_dir, lb_method, index_field, None  # TODO: Fix force vector
    )
    boundary_assignments = indexing.substitute_proxies(boundary_assignments)

    from pystencils.types.quick import SInt
    from pystencils import CreateKernelConfig, Assignment, create_kernel
    from pystencils.simp import add_subexpressions_for_field_reads

    config = CreateKernelConfig(
        index_field=index_field,
        target=target,
        index_dtype=SInt(32),
        default_dtype=pdf_field.dtype,
        skip_independence_check=True,
        **kernel_creation_args,
    )

    default_data_type = config.default_dtype
    if pdf_field.dtype != default_data_type:
        boundary_assignments = add_subexpressions_for_field_reads(
            boundary_assignments, data_type=default_data_type
        )

    elements: list[Assignment] = []

    index_arrs_node = indexing.create_code_node()
    elements += index_arrs_node.get_array_declarations()

    for node in boundary_functor.get_additional_code_nodes(lb_method)[::-1]:
        elements += node.get_array_declarations()

    elements += [Assignment(dir_symbol, index_field[0]("dir"))]
    elements += boundary_assignments.all_assignments

    asms = AssignmentCollection(elements)
    asms = BlockforestParameters.process(asms)

    kernel = create_kernel(asms, config=config)
    return kernel
