from __future__ import annotations

from dataclasses import dataclass

from pystencils import CreateKernelConfig
from pystencils.types.quick import Fp
from pystencils.jit import no_jit

from pystencilssfg import SfgContext
from pystencilssfg.composer import SfgIComposer


def cmake_parse_bool(var: str):
    var = var.upper()
    if var in ("ON", "1", "TRUE"):
        return True
    elif var in ("OFF", "0", "FALSE"):
        return False
    else:
        raise ValueError(f"Could not parse cmake value `{var}` as boolean.")


@dataclass
class WalberlaBuildConfig:
    """Represents a waLBerla build system configuration"""

    c_compiler_id: str
    """Value of `CMAKE_C_COMPILER_ID`."""

    cxx_compiler_id: str
    """Value of `CMAKE_CXX_COMPILER_ID`."""

    use_double_precision: bool
    """Value of `WALBERLA_DOUBLE_ACCURACY`"""

    optimize_for_localhost: bool
    """Value of `WALBERLA_OPTIMIZE_FOR_LOCALHOST`."""

    mpi_enabled: bool
    """Value of `WALBERLA_BUILD_WITH_MPI`."""

    openmp_enabled: bool
    """Value of `WALBERLA_BUILD_WITH_OPENMP`."""

    cuda_enabled: bool
    """Value of `WALBERLA_BUILD_WITH_CUDA`."""

    hip_enabled: bool
    """Value of `WALBERLA_BUILD_WITH_HIP`."""

    likwid_enabled: bool
    """Value of `WALBERLA_BUILD_WITH_LIKWID_MARKERS`"""

    @staticmethod
    def from_sfg(sfg: SfgContext | SfgIComposer) -> WalberlaBuildConfig:
        if isinstance(sfg, SfgIComposer):
            ctx = sfg.context
        else:
            ctx = sfg

        if isinstance(ctx.project_info, WalberlaBuildConfig):
            return ctx.project_info
        else:
            raise ValueError(
                "The given SfgContext does not encapsulate a waLBerla build config object."
            )

    def get_pystencils_config(self) -> CreateKernelConfig:
        dtype = Fp(64) if self.use_double_precision else Fp(32)

        return CreateKernelConfig(
            default_dtype=dtype,
            jit=no_jit,
        )
