from typing import Iterable
import sympy as sp

from pystencils.types import deconstify
from pystencilssfg.ir.postprocessing import SfgDeferredNode, PostProcessingContext
from pystencilssfg.ir import (
    SfgCallTreeNode,
    SfgStatements,
    SfgVisibility,
    SfgVisibilityBlock,
    SfgMemberVariable,
    SfgSequence,
    SfgConstructor,
)
from pystencilssfg.lang import AugExpr, SfgVar
from pystencilssfg.exceptions import SfgException


class CaptureToClass(SfgDeferredNode):
    def __init__(
        self, ignore: Iterable[str | SfgVar | AugExpr | sp.Symbol] = ()
    ) -> None:
        self._ignore = set()
        for ig in ignore:
            if isinstance(ig, AugExpr):
                self._ignore |= ig.depends
            elif isinstance(ig, sp.Symbol):
                self._ignore.add(ig.name)
            else:
                self._ignore.add(ig)

    def expand(self, ppc: PostProcessingContext) -> SfgCallTreeNode:
        cls = ppc.enclosing_class
        if cls is None:
            raise SfgException("Cannot capture variables to class in a free function")

        statements: list[SfgStatements] = []
        captured: list[SfgVar] = []
        members_block = SfgVisibilityBlock(SfgVisibility.PRIVATE)

        for var in ppc.live_variables:
            if not (var in self._ignore or var.name in self._ignore):
                member = SfgMemberVariable(f"{var}_", deconstify(var.dtype))
                members_block.append_member(member)
                captured.append(var)
                statements.append(
                    SfgStatements(f"{var.dtype.c_string()} {var.name} {{ {member} }};", [var], [])
                )

        cls.append_visibility_block(members_block)

        constructor_block = SfgVisibilityBlock(SfgVisibility.PUBLIC)
        constructor_block.append_member(
            SfgConstructor(captured, [f"{var}_ {{ {var} }}" for var in captured])
        )
        cls.append_visibility_block(constructor_block)

        return SfgSequence(statements)
