from typing import Sequence, Iterable
from dataclasses import dataclass, replace
from itertools import chain
from collections import defaultdict
import warnings

import sympy as sp

from pystencilssfg.composer.custom import CustomGenerator
from pystencilssfg.composer.class_composer import SfgClassComposer
from pystencilssfg.ir import SfgMethod

from pystencils import (
    Assignment,
    AssignmentCollection,
    CreateKernelConfig,
    Field,
    FieldType,
    Target,
    create_type,
)
from pystencils.types import PsType, constify, deconstify, PsCustomType, PsStructType

from pystencilssfg import SfgComposer
from pystencilssfg.lang import (
    VarLike,
    asvar,
    SfgVar,
    AugExpr,
    Ref,
    SrcVector,
    strip_ptr_ref,
)
from pystencilssfg.lang.types import CppTypeFactory, cpptype
from .reflection import GeneratedClassWrapperBase
from .api import (
    StructuredBlockForest,
    GenericWalberlaField,
    GhostLayerFieldPtr,
    GpuFieldPtr,
    IndexListBufferPtr,
    IBlockPtr,
    BlockDataID,
    Vector2,
    Vector3,
    CellInterval,
    CellIdx,
)


class SweepClassProperties:
    @dataclass(frozen=True)
    class Property:
        name: str
        dtype: PsType
        getter: bool
        setter: bool
        initializer: tuple[AugExpr, ...] | None = None

        def __post_init__(self):
            if self.dtype.const and self.setter:
                raise ValueError("Constant properties cannot have setters.")

        @property
        def var(self) -> SfgVar:
            return SfgVar(self.name, self.dtype)

        @property
        def member_var(self) -> SfgVar:
            return SfgVar(self.name + "_", self.dtype)

        def make_methods(self, sfg: SfgComposer) -> Sequence[SfgMethod]:
            methods = []

            if self.getter:
                methods.append(
                    sfg.method(
                        f"{self.name}",
                        returns=Ref(constify(self.dtype)),
                        const=True,
                        inline=True,
                    )(f"return {self.name}_;")
                )

            if self.setter:
                methods.append(
                    sfg.method(
                        f"{self.name}",
                        returns=Ref(self.dtype),
                        const=False,
                        inline=True,
                    )(f"return {self.name}_;")
                )

            return methods

    def __init__(self):
        self._properties: dict[str, SweepClassProperties.Property] = dict()

    @property
    def properties(self) -> Iterable[Property]:
        return self._properties.values()

    def add_property(
        self,
        prop: VarLike,
        const=False,
        setter: bool = True,
        getter: bool = True,
        initializer: tuple[AugExpr, ...] | None = None,
    ):
        prop = asvar(prop)
        if prop.name in self._properties:
            raise ValueError(f"Duplicate property: {prop}")

        dtype = constify(prop.dtype) if const else deconstify(prop.dtype)

        self._properties[prop.name] = SweepClassProperties.Property(
            prop.name, dtype, getter, setter, initializer
        )

    def get(self, prop: VarLike) -> AugExpr:
        prop = asvar(prop)

        if prop.name not in self._properties:
            raise KeyError(f"Property not registered: {prop}")

        return AugExpr.format("this->{}_", prop.name)

    def render(self, sfg: SfgComposer):
        priv = sfg.private(*(p.member_var for p in self.properties))

        ctor = self._constructor(sfg)

        for p in self.properties:
            if p.initializer is None:
                ctor.add_param(p.var)
            else:
                for var in chain.from_iterable(e.depends for e in p.initializer):
                    if var not in ctor.parameters:
                        if (
                            deconstify(strip_ptr_ref(var.dtype))
                            == StructuredBlockForest.typename()
                        ):
                            ctor.add_param(var, 0)
                        else:
                            ctor.add_param(var)

        for p in self._properties.values():
            if p.initializer is not None:
                ctor.init(p.member_var)(*p.initializer)
            else:
                ctor.init(p.member_var)(p.var)

        methods = chain.from_iterable(p.make_methods(sfg) for p in self.properties)
        pub = sfg.public(ctor, *(methods))

        return priv, pub

    def _constructor(self, sfg: SfgComposer) -> SfgClassComposer.ConstructorBuilder:
        return sfg.constructor()


class BlockforestParameters:
    def __init__(
        self,
        property_cache: SweepClassProperties,
        block: IBlockPtr,
        cell_interval: AugExpr | None = None,
    ):
        self._property_cache = property_cache
        self._block = block
        self._ci = cell_interval
        self._extractions: dict[SfgVar, AugExpr] = dict()

        self._blockforest: StructuredBlockForest | None = None

    @staticmethod
    def process(asms: AssignmentCollection) -> AssignmentCollection:
        from sympy.core.function import AppliedUndef

        expandable_appls: set[AppliedUndef] = filter(  # type: ignore
            lambda expr: hasattr(expr, "expansion_func"), asms.atoms(AppliedUndef)
        )

        subs: dict[AppliedUndef, sp.Symbol] = dict()
        for appl in expandable_appls:
            expansion: sp.Expr = appl.expansion_func(*appl.args)  # type: ignore
            symb = next(asms.subexpression_symbol_generator)
            asms.subexpressions.insert(0, Assignment(symb, expansion))
            subs[appl] = symb

        return asms.new_with_substitutions(subs)

    def blockforest(self) -> StructuredBlockForest:
        if self._blockforest is None:
            blocks = StructuredBlockForest.shared_ptr()
            self._property_cache.add_property(
                blocks,
                getter=False,
                setter=False,
                initializer=(StructuredBlockForest.shared_ptr_ref(),),
            )
            self._blockforest = StructuredBlockForest().bind(
                "(*{})", self._property_cache.get(blocks)
            )
            return self._blockforest
        else:
            return self._blockforest

    def filter_params(self, params: set[SfgVar]) -> set[SfgVar]:
        from .symbolic import domain, block, cell

        params_filtered = set()

        for param in params:
            for coord in range(3):
                if param.name == domain.aabb_min[coord].name:
                    self._extractions[param] = (
                        self.blockforest().getDomain().min()[coord]
                    )
                    break
                elif param.name == domain.aabb_max[coord].name:
                    self._extractions[param] = (
                        self.blockforest().getDomain().max()[coord]
                    )
                    break
                elif param.name == block.aabb_min[coord].name:
                    self._extractions[param] = self._block.getAABB().min()[coord]
                    break
                elif param.name == block.aabb_max[coord].name:
                    self._extractions[param] = self._block.getAABB().max()[coord]
                    break
                elif param.name == block.ci_min[coord].name:
                    if self._ci is None:
                        self._extractions[param] = AugExpr.format("0")
                    else:
                        raise NotImplementedError()
                    break
                elif param.name == block.ci_max[coord].name:
                    if self._ci is None:
                        self._extractions[param] = AugExpr.format(
                            "{} - 1",
                            self.blockforest().cellsPerBlock(coord, self._block),
                        )
                    else:
                        raise NotImplementedError()
                    break
                elif param.name == cell.cell_extents[coord].name:
                    self._extractions[param] = self.blockforest().cell_extents(
                        coord, self.blockforest().getLevel(self._block.deref())
                    )
                    break
            else:
                params_filtered.add(param)

        return params_filtered

    def render_extractions(self, sfg: SfgComposer):
        return (sfg.init(p)(expr) for p, expr in self._extractions.items())


@dataclass
class FieldInfo:
    field: Field
    wlb_field_ptr: GenericWalberlaField | IndexListBufferPtr
    data_id: BlockDataID

    def as_glfield(self) -> GenericWalberlaField:
        if not isinstance(self.wlb_field_ptr, GenericWalberlaField):
            raise AttributeError(
                "This FieldInfo does not encapsulate a ghost layer field"
            )
        return self.wlb_field_ptr

    def is_glfield(self) -> bool:
        return isinstance(self.wlb_field_ptr, GenericWalberlaField)

    def as_index_list(self) -> IndexListBufferPtr:
        if not isinstance(self.wlb_field_ptr, IndexListBufferPtr):
            raise AttributeError("This FieldInfo does not encapsulate an index list")
        return self.wlb_field_ptr

    def is_index_list(self) -> bool:
        return isinstance(self.wlb_field_ptr, IndexListBufferPtr)


class ShadowFieldCache:
    """
    Information needed:
        - Name and data type of field
        - Name of shadow field
    """

    def __init__(self):
        self._originals: dict[str, FieldInfo] = dict()

    def add_shadow(self, original: FieldInfo):
        self._originals[original.field.name] = original

    def get_shadow(self, original_name: str) -> AugExpr:
        original = self._originals[original_name]
        return AugExpr.format(
            "this->{}({})", self._getter(original.field.name), original.wlb_field_ptr
        )

    def perform_swap(self, original_name: str, shadow_field: FieldInfo) -> AugExpr:
        original = self._originals[original_name]
        return original.as_glfield().swapDataPointers(shadow_field.wlb_field_ptr)

    def render(self, sfg: SfgComposer):
        if not self._originals:
            return ()

        sfg.include("<memory>")

        cache_ptrs = []
        getters = []

        for orig_name, orig in self._originals.items():
            unique_ptr_type = PsCustomType(
                f"std::unique_ptr< {orig.as_glfield().field_type} >"
            )
            cache_ptr_name = self._cache_ptr(orig.as_glfield())
            cache_ptrs.append(sfg.var(cache_ptr_name, unique_ptr_type))

            getters.append(
                sfg.method(
                    self._getter(orig_name),
                    returns=orig.wlb_field_ptr.get_dtype(),
                    inline=True,
                )(
                    sfg.branch(f"{cache_ptr_name} == nullptr")(
                        AugExpr.format(
                            "{}.reset({});",
                            cache_ptr_name,
                            orig.as_glfield().cloneUninitialized(),
                        )
                    ),
                    f"return {cache_ptr_name}.get();",
                )
            )

        return (sfg.private(*cache_ptrs, *getters),)

    def _getter(self, orig_name: str) -> str:
        return f"_getShadow_{orig_name}"

    def _cache_ptr(self, orig: GenericWalberlaField) -> str:
        return f"shadow_{str(orig)}_"


def combine_vectors(scalars: set[SfgVar]) -> dict[SrcVector, tuple[SfgVar, ...]]:
    """Attempt to combine vector component symbols into vectors.

    This function modifies the `scalars` parameter in-place by removing grouped scalar
    components and replacing them by their vectors.
    """
    potential_vectors: defaultdict[str, list[tuple[int, SfgVar]]] = defaultdict(list)

    for s in scalars:
        tokens = s.name.split("_")
        try:
            coord = int(tokens[-1])
            vec_name = "_".join(tokens[:-1])
            potential_vectors[vec_name].append((coord, s))
        except ValueError:
            pass

    all_vectors: dict[SrcVector, tuple[SfgVar, ...]] = dict()

    for name, entries in potential_vectors.items():
        entries = sorted(entries, key=lambda t: t[0])
        coords = [e[0] for e in entries]
        components = tuple(e[1] for e in entries)
        scalar_types = set(e[1].dtype for e in entries)

        if set(coords) == {0, 1}:
            vec_type = Vector2
        elif set(coords) == {0, 1, 2}:
            vec_type = Vector3
        else:
            continue

        if len(scalar_types) > 1:
            warnings.warn(
                f"Skipping vector grouping for vector {name}: "
                f"Encountered vector components {components} with conflicting data types {scalar_types}.",
                UserWarning,
            )
            continue

        name_conflict = False
        for s in scalars:
            if s.name == name:
                warnings.warn(
                    f"Skipping vector grouping for vector {name}: "
                    f"Vector name conflicts with existing scalar parameter {s}.",
                    UserWarning,
                )
                name_conflict = True

        if name_conflict:
            continue

        scalar_type = deconstify(scalar_types.pop())
        vector = vec_type(scalar_type).var(name)
        all_vectors[vector] = components

        for c in components:
            scalars.remove(c)
        scalars.add(asvar(vector))

    return all_vectors


class Sweep(CustomGenerator):
    """Generate a waLBerla sweep from a pystencils kernel."""

    def __init__(
        self,
        name: str,
        assignments: Sequence[Assignment] | AssignmentCollection,
        config: CreateKernelConfig | None = None,
    ):
        if config is not None:
            config = config.copy()

            if config.get_option("ghost_layers") is not None:
                raise ValueError(
                    "Specifying `ghost_layers` in your codegen config is invalid when generating a waLBerla sweep."
                )

            if (
                config.get_option("iteration_slice") is None
                and config.get_option("index_field") is None
            ):
                config.ghost_layers = 0
        else:
            config = CreateKernelConfig(ghost_layers=0)

        self._name = name

        if isinstance(assignments, AssignmentCollection):
            self._assignments = assignments
        else:
            self._assignments = AssignmentCollection(assignments)  # type: ignore
        self._gen_config = config

        if self._gen_config.get_target() == Target.CUDA:
            self._glfield_type = GpuFieldPtr
        elif self._gen_config.get_target().is_cpu():
            self._glfield_type = GhostLayerFieldPtr
        else:
            raise ValueError(
                f"Cannot generate sweep for target {self._gen_config.target}"
            )

        #   Map from shadow field to shadowed field
        self._shadow_fields: dict[Field, Field] = dict()

        #   RESULTS - unset at this point
        self._generated_class: type[GeneratedClassWrapperBase] | None = None

    #   CONFIGURATION

    @property
    def sparse(self) -> bool:
        return self._gen_config.get_option("index_field") is not None

    @sparse.setter
    def sparse(self, sparse_iteration: bool):
        if sparse_iteration:
            if self._gen_config.get_option("index_field") is not None:
                return

            if self._gen_config.get_option("index_dtype") != create_type("int64"):
                raise ValueError(
                    "Sparse sweeps require `int64_t` as their index data type. Check your code generator config."
                )

            self._gen_config.index_field = Field.create_generic(
                "indexList", 1, CellIdx, field_type=FieldType.INDEXED
            )
            self._gen_config.ghost_layers = None
        else:
            self._gen_config.index_field = None
            self._gen_config.ghost_layers = 0

    def swap_fields(self, field: Field, shadow_field: Field):
        if field in self._shadow_fields:
            raise ValueError(f"Field swap for {field} was already registered.")
        if shadow_field in self._shadow_fields:
            raise ValueError(f"Field swap for {shadow_field} was already registered.")
        if field.dtype != shadow_field.dtype:
            raise ValueError(
                "Field and its shadow must have the same element type for swapping"
            )

        self._shadow_fields[shadow_field] = field

    #   CODE GENERATION

    def _walberla_field(self, f: Field) -> GenericWalberlaField | IndexListBufferPtr:
        match f.field_type:
            case FieldType.GENERIC | FieldType.CUSTOM:
                return self._glfield_type.create(f)
            case FieldType.INDEXED:
                assert isinstance(f.dtype, PsStructType)
                return IndexListBufferPtr(f.dtype).var(f.name)
            case _:
                raise ValueError(
                    f"Unable to map field {f} of type {f.field_type} to a waLBerla field."
                )

    def generate(self, sfg: SfgComposer) -> None:
        knamespace = sfg.kernel_namespace(f"{self._name}_kernels")

        assignments = BlockforestParameters.process(self._assignments)
        #   TODO: Get default config from waLBerla build system and override its entries
        #   from the user-provided config
        khandle = knamespace.create(assignments, self._name, self._gen_config)

        all_fields: dict[str, FieldInfo] = {
            f.name: FieldInfo(
                f, self._walberla_field(f), BlockDataID().var(f"{f.name}Id")
            )
            for f in khandle.fields
        }

        swaps: dict[str, FieldInfo] = dict()
        shadows_cache = ShadowFieldCache()

        for shadow, orig in self._shadow_fields.items():
            swaps[orig.name] = all_fields[shadow.name]
            shadows_cache.add_shadow(all_fields[orig.name])

        block = IBlockPtr().var("block")
        block_fields = sorted(
            (fi for fi in all_fields.values() if fi.field not in self._shadow_fields),
            key=lambda fi: fi.field.name,
        )

        props = SweepClassProperties()

        parameters = khandle.scalar_parameters

        blockforest_params = BlockforestParameters(props, block, None)
        parameters = blockforest_params.filter_params(parameters)

        vector_groups = combine_vectors(parameters)

        for fi in block_fields:
            props.add_property(fi.data_id, setter=False, getter=True)

        for s in sorted(parameters, key=lambda p: p.name):
            props.add_property(s, setter=True, getter=True)

        def render_runmethod(ci: CellInterval | None = None):
            return [
                #   Get IDs from class
                *(sfg.init(fi.data_id)(props.get(fi.data_id)) for fi in block_fields),
                #   Extract regular fields from block
                *(
                    sfg.init(fi.wlb_field_ptr)(
                        block.getData(fi.wlb_field_ptr.field_type, fi.data_id)
                    )
                    for fi in block_fields
                ),
                #   Get shadow fields from cache
                *(
                    sfg.init(shadow_info.wlb_field_ptr)(
                        shadows_cache.get_shadow(orig_name)
                    )
                    for orig_name, shadow_info in swaps.items()
                ),
                #   Map GhostLayerFields to pystencils fields
                *(
                    sfg.map_field(fi.field, fi.as_glfield().with_cell_interval(ci))
                    for fi in all_fields.values()
                    if fi.is_glfield()
                ),
                #   Map index lists to pystencils fields
                *(
                    sfg.map_field(fi.field, fi.as_index_list())
                    for fi in all_fields.values()
                    if fi.is_index_list()
                ),
                #   Get parameters from class
                *(sfg.init(param)(props.get(param)) for param in parameters),
                #   Extract scalars from vectors
                *(
                    sfg.map_vector(components, vector)
                    for vector, components in vector_groups.items()
                ),
                #   Extract geometry information
                *(blockforest_params.render_extractions(sfg)),
                sfg.call(khandle),
                #   Perform field swaps
                *(
                    shadows_cache.perform_swap(orig_name, shadow_info)
                    for orig_name, shadow_info in swaps.items()
                ),
            ]

        methods = [sfg.method("operator()")(*render_runmethod())]

        if not self.sparse:
            methods.append(
                sfg.method("runOnCellInterval")(
                    *render_runmethod(CellInterval(const=True, ref=True).var("ci"))
                )
            )

        sfg.klass(self._name)(
            *props.render(sfg),
            sfg.public(*methods),
            *shadows_cache.render(sfg),
        )

        gen_class = sfg.context.get_class(self._name)
        assert gen_class is not None

        class GenClassWrapper(GeneratedClassWrapperBase):
            _class = gen_class
            _namespace = sfg.context.fully_qualified_namespace

        self._generated_class = GenClassWrapper

    #   CODE-GENERATION RESULTS
    #   These properties are only available after the sweep was generated

    @property
    def generated_class(self) -> type[GeneratedClassWrapperBase]:
        if self._generated_class is None:
            raise AttributeError(
                "Generated class is unavailable - code generation was not run yet."
            )

        return self._generated_class
