
#include "core/all.h"
#include "blockforest/all.h"

#include "field/all.h"
#include "field/communication/PackInfo.h"
#include "field/communication/StencilRestrictedPackInfo.h"

#include "geometry/all.h"

#include "vtk/all.h"

#include <array>

#include "gen/IrregularFreeSlip.hpp"

namespace TestIrregularFreeSlip
{
    using namespace walberla;

    using PdfField_T = field::GhostLayerField<real_t, gen::LbStencil::Q>;
    using ScalarField_T = field::GhostLayerField<real_t, 1>;
    using VectorField_T = field::GhostLayerField<real_t, gen::LbStencil::D>;
    using FlagField_T = FlagField<uint8_t>;
    using CommScheme = blockforest::communication::UniformBufferedScheme<gen::LbStencil>;
    using PdfsPackInfo = field::communication::StencilRestrictedPackInfo<PdfField_T, gen::LbStencil>;

    /**
     * Pipe flow with circular cross-section and free-slip walls.
     * The pipe flow is initialized with a constant velocity in x-direction.
     * As free-slip walls do not impose any friction, velocity should remain constant in time.
     */
    void freeSlipPipe(Environment &env)
    {
        std::array<uint_t, 3> blocks{1, 1, 1};
        std::array<uint_t, 3> cpb{4, 32, 32};
        std::array<bool, 3> periodic{true, false, false};

        auto sbfs = blockforest::createUniformBlockGrid(
            blocks[0], blocks[1], blocks[2],
            cpb[0], cpb[1], cpb[2],
            1.0,
            true,
            periodic[0], periodic[1], periodic[2]
        );

        BlockDataID pdfsId = field::addToStorage<PdfField_T>(sbfs, "f", real_c(0.0));
        BlockDataID rhoId = field::addToStorage<ScalarField_T>(sbfs, "rho", real_c(0.0));
        BlockDataID uId = field::addToStorage<VectorField_T>(sbfs, "u", real_c(0.0));

        const BlockDataID flagFieldId = field::addFlagFieldToStorage<FlagField_T>(sbfs, "flagField");
        const FlagUID fluidFlagUid{"Fluid"};
        const FlagUID freeSlipFlagUID{"FreeSlip"};

        const CellInterval allCells{{0, 0, 0}, {sbfs->getNumberOfXCellsPerBlock() - 1, sbfs->getNumberOfYCellsPerBlock() - 1, sbfs->getNumberOfZCellsPerBlock() - 1}};
        const CellInterval allCellsWithGl{{-1, 0, 0}, {sbfs->getNumberOfXCellsPerBlock(), sbfs->getNumberOfYCellsPerBlock() - 1, sbfs->getNumberOfZCellsPerBlock() - 1}};

        const real_t pipeRadius{14.0};
        const Vector3<real_t> pipeAnchor{0.0, 16.0, 16.0};
        const real_t maxVelocity{0.02};

        for (auto &block : *sbfs)
        {
            FlagField_T &flagField = *block.getData<FlagField_T>(flagFieldId);
            const uint8_t freeSlipFlag{flagField.getOrRegisterFlag(freeSlipFlagUID)};

            VectorField_T &velField = *block.getData<VectorField_T>(uId);

            for (Cell c : allCellsWithGl)
            {
                Cell globalCell{c};
                sbfs->transformBlockLocalToGlobalCell(globalCell, block);
                Vector3<real_t> cellCenter{sbfs->getCellCenter(globalCell)};
                cellCenter[0] = 0.0;

                Vector3<real_t> initVelocity;
                real_t radialDistance = (cellCenter - pipeAnchor).length();
                if (radialDistance > pipeRadius)
                {
                    flagField.addFlag(c, freeSlipFlag);
                    initVelocity = Vector3<real_t>{NAN};
                }
                else
                {
                    initVelocity = Vector3<real_t>{maxVelocity, 0., 0.};
                }

                velField.get(c, 0) = initVelocity[0];
                velField.get(c, 1) = initVelocity[1];
                velField.get(c, 2) = initVelocity[2];
            }
        }

        geometry::setNonBoundaryCellsToDomain<FlagField_T>(*sbfs, flagFieldId, fluidFlagUid);

        auto flagsOutput = field::createVTKOutput<FlagField_T>(flagFieldId, *sbfs, "flags");
        flagsOutput();

        gen::LbInit initialize { pdfsId, uId, 1.0 };

        for(auto& b : *sbfs){
            initialize(&b);
        }

        gen::FreeSlipIrregular freeSlip{
            gen::FreeSlipIrregularFactory(sbfs, pdfsId).fromFlagField<FlagField_T>(flagFieldId, freeSlipFlagUID, fluidFlagUid)
        };

        gen::LbStreamCollide streamCollide { pdfsId, rhoId, uId, 1.0 };

        CommScheme comm{sbfs};
        auto pdfsPackInfo = std::make_shared<PdfsPackInfo>(pdfsId);
        comm.addPackInfo(pdfsPackInfo);

        auto velOutput = field::createVTKOutput<VectorField_T>(uId, *sbfs, "vel");

        for(uint_t i = 0; i < 10; ++i){
            comm();
            for (auto &block : *sbfs)
            {
                velOutput();
                freeSlip(&block);
                streamCollide(&block);

                VectorField_T & velField = *block.getData< VectorField_T >(uId);
                FlagField_T & flagField = *block.getData< FlagField_T >(flagFieldId);
                uint8_t fluidFlag = flagField.getFlag(fluidFlagUid);
                
                for(Cell c : allCells){
                    if( flagField.isFlagSet(c, fluidFlag) ){
                        WALBERLA_CHECK_FLOAT_EQUAL( velField.get(c, 0), maxVelocity );
                        WALBERLA_CHECK_FLOAT_EQUAL( velField.get(c, 1), 0. );
                        WALBERLA_CHECK_FLOAT_EQUAL( velField.get(c, 2), 0. );
                    }
                }
            }
        }

        velOutput();
    }
}

int main(int argc, char **argv)
{
    walberla::Environment env{argc, argv};
    TestIrregularFreeSlip::freeSlipPipe(env);
}
