import sympy as sp
import pystencils as ps

from pystencilssfg import SourceFileGenerator
from sfg_walberla import Sweep
from sfg_walberla.boundaries import SimpleHbbBoundary

from lbmpy import (
    LBStencil,
    Stencil,
    Method,
    LBMConfig,
    LBMOptimisation,
    create_lb_method,
    create_lb_update_rule,
)

from lbmpy.boundaries import NoSlip
from lbmpy.macroscopic_value_kernels import macroscopic_values_setter

stencil = LBStencil(Stencil.D3Q19)
dim = stencil.D
f, f_tmp, rho, u = ps.fields(
    f"f({stencil.Q}), f_tmp({stencil.Q}), rho(1), u({dim}): double[{dim}D]", layout="fzyx"
)
omega = sp.Symbol("omega")
force = sp.symbols(f"F_:{dim}")

lbm_config = LBMConfig(
    stencil=stencil,
    method=Method.CENTRAL_MOMENT,
    relaxation_rate=omega,
    force=force,
    compressible=True,
    zero_centered=False,
    output={"density": rho, "velocity": u},
)

lb_method = create_lb_method(lbm_config)

with SourceFileGenerator() as sfg:
    sfg.namespace("Ex_ForceDrivenChannel::gen")

    sfg.include(f"stencil/{stencil.name}.h")
    sfg.code(f"using LbStencil = walberla::stencil::{stencil.name};")

    lbm_opt = LBMOptimisation(symbolic_field=f, symbolic_temporary_field=f_tmp)

    gen_config = ps.CreateKernelConfig()
    gen_config.target = ps.Target.CPU
    gen_config.cpu.openmp.enable = True

    lb_update = create_lb_update_rule(lbm_config=lbm_config, lbm_optimisation=lbm_opt)
    lb_update_sweep = Sweep("LbStreamCollide", lb_update, gen_config)
    lb_update_sweep.swap_fields(f, f_tmp)
    sfg.generate(lb_update_sweep)

    lb_init = macroscopic_values_setter(
        lb_update.method, density=rho, velocity=u, pdfs=f, set_pre_collision_pdfs=True
    )
    lb_init_sweep = Sweep("LbInit", lb_init, gen_config)
    sfg.generate(lb_init_sweep)

    #   No-Slip Wall
    sfg.generate(SimpleHbbBoundary(NoSlip(), lb_method, f))
